# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

from fingertip.plugins.script._run_and_test import run, test

# debug is in fingertip.plugins.script.debug,
# but I'm not importing to minimize dependencies / rebuilds
